/*
 * %Z%%M%    %I%  %W%  %G%  %U%
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 * IBM SDK, Java(tm) 2 Technology Edition, v5.0
 * (C) Copyright IBM Corp. 2004, 2005. All Rights Reserved
 * ===========================================================================
 */
import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;

import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.KeyManager;

/** Client/Server - ClientPKCS12KeyJsse
  * <P>
  * This example shows a simple <I>Client/Server interaction</I> using the new JSSE2 provider.
  * For the sake of simplicity the ClientPKCS12KeyJsse and the ServerPKCS12KeyJsse (see ServerPKCS12KeyJsse.java) use
  * the same test PKCS12 KeyStore file testkeys.pfx. In real life there must be two different
  * PKCS12 KeyStore files, one for the client, and one for the server.
  * The PKCS12 file testkeys.pfx is protected by the password "passphrase".
  * When the ServerPKCS12KeyJsse is started it waits for the ClientPKCS12KeyJsse
  * connection. The ServerPKCS12KeyJsse uses the port 8050. To start the ClientPKCS12KeyJsse the host name
  * of the ServerPKCS12KeyJsse must be supplied (default: localhost).
  * First, ClientPKCS12KeyJsse and ServerPKCS12KeyJsse transmit a number of messages (N round-trips)
  * over several SSL connections; ClientPKCS12KeyJsse tries to connect to ServerPKCS12KeyJsse using all the
  * enabled cipher suites.
  * Than they establish a normal TCP connection and transmit the same number of
  * messages over it.
  */
public class ClientPKCS12KeyJsse {
    static int N = 50, L = 100, R = 2;

    static SSLContext sslContext;

    ClientPKCS12KeyJsse() {
        super();
    }

    public static void initContext() {
        try {

            if (Security.getProvider("com.ibm.jsse2.IBMJSSEProvider2") == null) {
                // IBMJSSE2 Provider is not installed, install it
                Security.addProvider(new com.ibm.jsse2.IBMJSSEProvider2());
            }

            KeyStore ks = KeyStore.getInstance("pkcs12");
            ks.load(new FileInputStream("testkeys.pfx"), "passphrase".toCharArray());
            KeyManagerFactory kmf = KeyManagerFactory.getInstance("NewIbmX509","IBMJSSE2");
            kmf.init(ks, "passphrase".toCharArray());

            TrustManager[] tm;
            TrustManagerFactory tmf = TrustManagerFactory.getInstance("PKIX","IBMJSSE2");
            tmf.init(ks);
            tm = tmf.getTrustManagers();

            sslContext = SSLContext.getInstance("SSL_TLS","IBMJSSE2");
            sslContext.init(kmf.getKeyManagers(), tm, null);
        } catch (Throwable t) {
            System.out.println("Failed to read testkeys.pfx file.");
            t.printStackTrace();
        }
    }


/** Starts the ClientPKCS12KeyJsse application.
  * @param args the name of the host the ClientPKCS12KeyJsse should connect to
  */
    public static void main(String args[]) {

        int i, j, len, ci_nonanon;

        String host = "127.0.0.1";
        int port = 8050;

        if ((args != null) && (args.length > 0)) {
            host = args[0];
            i = host.indexOf(':');
            if (i != -1) {
                try {
                    port = Integer.parseInt(host.substring(i+1));
                    host = host.substring(0,i);
                } catch (Exception e) {
                    System.err.println("ClientPKCS12KeyJsse: wrong address format");
                    System.exit(1);
                }
            }
        }
        if (args.length > 1)
            N = Integer.parseInt(args[1]);
        if (args.length > 2)
            L = Integer.parseInt(args[2]);
        byte buffer[] = new byte[L];


        System.out.println("ClientPKCS12KeyJsse: started.");

        initContext();


        try {
            if (sslContext == null) {
                System.out.println ("ClientPKCS12KeyJsse:  SSL client context NOT created.");
                return;
            }
            System.out.println("ClientPKCS12KeyJsse: SSL server context created.");
            SSLSocketFactory factory = sslContext.getSocketFactory();
            String[] cipher_suites = sslContext.getSocketFactory().getSupportedCipherSuites();

            System.out.println("ClientPKCS12KeyJsse: enabled cipher suites");
            for (i=0, ci_nonanon=0; i<cipher_suites.length; i++) {
                if ((cipher_suites[i].indexOf("_anon_") < 0)  && (cipher_suites[i].indexOf("KRB5") < 0)){
                    ci_nonanon++;
                }
                System.out.println("   " + cipher_suites[i]);
            }

            SSLSocket ssl_sock;
            OutputStream ostr;
            InputStream istr;
            long t;
            String[] ecs = new String[1];

            for (int csi = 0; csi<ci_nonanon; csi++) {
                // Remove anonymous cipher suites. TrustManager requires a personal certificate.
                ecs[0] = cipher_suites[csi];
                if ((ecs[0].indexOf("_anon_") >= 0) || (ecs[0].indexOf("KRB5") >= 0)) {
                    continue;
                }
                for (int n = 0; n < R; n++) {
                    t = System.currentTimeMillis();
                    try {
                        ssl_sock = (SSLSocket) factory.createSocket(host, port);
                        ssl_sock.setEnabledCipherSuites(ecs);

                        ssl_sock.startHandshake();
                        SSLSession session = ssl_sock.getSession();
                        System.out.println("\nClientPKCS12KeyJsse: SSL connection established");
                        System.out.println("   cipher suite:       " + session.getCipherSuite());
                    } catch (Exception se) {
                        System.out.println("\nClientPKCS12KeyJsse: can't connect using: " +
                                           cipher_suites[csi] + "\n" + se);
                        se.printStackTrace();
                        break;
                    }

                    if (L > 0) {
                        istr = ssl_sock.getInputStream();
                        ostr = ssl_sock.getOutputStream();

                        for (j=0; j<N; j++) {
                            if ((j==N-1) && (n == R-1) && (csi == ci_nonanon - 1))
                                buffer[0] = (byte)-1;
                            ostr.write(buffer, 0, L);
                            for (len = 0;;) {
                                try {
                                    if ((i = istr.read(buffer, len, L-len)) == -1) {
                                        System.out.println("ClientPKCS12KeyJsse: connection dropped by partner.");
                                        istr.close();
                                        return;
                                    }
                                    // System.out.println("ClientPKCS12KeyJsse: " + i + " bytes received.");
                                    if ((len+=i) == L) break;
                                } catch (InterruptedIOException e) {
                                    // System.out.println("ClientPKCS12KeyJsse: timeout.\n");
                                }
                            }
                        }
                    }
                    System.out.println("Messages = " + N*2 + "; Time = " +
                                       (System.currentTimeMillis() - t));
                    ssl_sock.close();
                    System.out.println("\nClientPKCS12KeyJsse: SSL connection closed.");
                }
            }

            Thread.sleep(1500);

            // plain socket benchmark
            if (L > 0) {
                // Socket bench
                Socket sock = new Socket(host, port);
                System.out.println("\nClientPKCS12KeyJsse: Socket connection established.");
                istr = sock.getInputStream();
                ostr = sock.getOutputStream();
                t = System.currentTimeMillis();
                for (j=0; j<N; j++) {
                    ostr.write(buffer, 0, L);
                    for (len = 0;;) {
                        if ((i = istr.read(buffer, len, L-len)) == -1) {
                            System.out.println("ClientPKCS12KeyJsse: connection dropped by partner.");
                            return;
                        }
                        // System.out.println("ClientPKCS12KeyJsse: " + i + " bytes received.");
                        if ((len+=i) == L) break;
                    }
                }
                System.out.println("Messages = " + N*2 + "; Time = " +
                                   (System.currentTimeMillis() - t));
                sock.close();
                System.out.println("\nClientPKCS12KeyJsse: Socket connection closed.");
                Thread.sleep(1500);
            }
        } catch (Exception e) {
            System.out.println(e);
        }
        System.out.println("ClientPKCS12KeyJsse: terminated.");
    }
}
