/*
 * %Z%%M%    %I%  %W%  %G%  %U%
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * IBM Security Software Development Kit, Java (tm) Technology Edition
 *
 * (C) Copyright IBM Corp. 2005 All Rights Reserved.
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 * ===========================================================================
 */

import java.io.*;
import java.net.*;
import java.nio.*; 
import java.nio.channels.*;
import java.security.*;
import java.util.*;
import javax.net.ssl.*;
import javax.net.ssl.SSLEngineResult.*;

/**
 * SSLClientEngine
 *
 * Sample application to demonstrate a simple client/server interaction using SSLEngine. 
 * 
 * This sample program uses the JSSE default cipher suites to secure the communications.
 * It uses the JKS file at the same directory called "testkeys" protected by the password 
 * "passphrase" as key store and trust store. At program start time, it loads the keys 
 * and certificates from the file to the memory, opens up a SocketChannel and initialize 
 * an SSLEngine. 
 * The engine is responsible for handling the exchange of handshake and application data messages 
 * represented in ByteBuffer. Data transportation is handled by SocketChannel. For simplicity 
 * this program uses SocketChannel in blocking mode. Users can always explore the NIO 
 * (in java.nio etc package) functionalities and utilize non-blocking mode to build more 
 * sophisticated applications. It sends the server a message, and receives a receipt 
 * from the server. After that, it closes the outbound channel and sends a close notify
 * message to the server. The inbound channel is closed when it receives the close
 * notify message from the server.
 *
 * Running the client/server application requires that the Server is started first and then the 
 * Client application.
 *
 * To start the Server program, execute the following command:
 *
 *     java SSLServerEngine
 *     or
 *     java SSLServerEngine <port number>
 *
 * The SSLServerEngine uses port 8050. Once it has started it waits for the client connection.
 *
 * To start the client, issue the following command:
 * 
 *     java SSLClientEngine <hostname> <port number>
 *     or
 *     java SSLClientEngine <hostname>
 *
 * where <hostname> is the name of the host running SSLServerEngine.  The default is localhost.
 * <port number> default is 8050. Both arguments are optional.
 *
 */

public class SSLClientEngine {
    private String name;
    private SSLContext context;
    private String host;
    private int port;
    private SocketChannel sc;

    SSLClientEngine(String name, String host, int port) {
        this.name = name;
        this.host = host;
        this.port = port;
    }

    private void initContext(String keystore, char[] password) {
        try {
            KeyStore ks = KeyStore.getInstance("JKS");
            ks.load(new FileInputStream(keystore), password);  //load the keys and certificates
            KeyManagerFactory kmf = KeyManagerFactory.getInstance("IbmX509");
            kmf.init(ks, password);  //initialize the key manager factory
            TrustManagerFactory tmf =  TrustManagerFactory.getInstance("PKIX");
            tmf.init(ks); //initialize the trust manager factory
            context = SSLContext.getInstance("SSL_TLS");
            context.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null); //initialized the ssl context with key manager and trust manager
        } catch (IOException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (KeyStoreException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (NoSuchAlgorithmException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (Exception e) {
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }

    private void initSocketChannel() throws IOException {
        InetSocketAddress sockAddress = new InetSocketAddress(host, port);
        sc = SocketChannel.open(sockAddress);
     }

    public static void main(String args[]) {
        String host = "127.0.0.1";
        int port = 8050;
        if (args.length == 1) {
            host = args[0];
        }
        else if (args.length == 2) {
            host = args[0];
            port = Integer.parseInt(args[1]);
        }
        
        SSLClientEngine client = new SSLClientEngine("CLIENT: ", host, port);

        System.out.println(client.name + "started.");
        client.initContext("testkeys", "passphrase".toCharArray());
        System.out.println(client.name + "connecting to " + host + ":" + port + " over SSL");
        try {
            client.initSocketChannel();
            client.run();
        }
        catch (IOException ex) {
            System.out.println(ex.getMessage());
            System.exit(1);
        }
    }

    private void run() throws IOException {
        SSLEngine engine = initEngine();
        int netBBSize = engine.getSession().getPacketBufferSize();
        int appBBSize = engine.getSession().getApplicationBufferSize();
        //allocate the max buffer size
        ByteBuffer netIn = ByteBuffer.allocate(netBBSize);
        ByteBuffer netOut = ByteBuffer.allocate(netBBSize);
        ByteBuffer appIn = ByteBuffer.allocate(appBBSize);
        ByteBuffer appOut = ByteBuffer.allocate(appBBSize);
        byte[] data = createMessage();
        doHandshake(engine, netIn, netOut, appIn, appOut, sc);
        processMessage(engine, netIn, netOut, appIn, appOut, sc, data);
        closeEngine(engine, netIn, netOut, appIn, appOut, sc);
    }

    private void processMessage(SSLEngine engine, ByteBuffer netIn,
                                ByteBuffer netOut, ByteBuffer appIn, 
                                ByteBuffer appOut, SocketChannel sc,
                                byte[] data) throws IOException {
        appOut.put(data);
        appOut.flip();
        SSLEngineResult result = engine.wrap(appOut, netOut);
        if (result.getStatus() == Status.OK) {
            System.out.println(name + "send message");
            netOut.flip();
            sc.write(netOut);
        }
        sc.read(netIn);
        netIn.flip();
        result = engine.unwrap(netIn, appIn);
        System.out.println(name + "server says");
        byte[] temp = new byte[appIn.flip().remaining()];
        appIn.get(temp);
        System.out.println(new String(temp));
        reset(appIn, appOut, netIn, netOut);
    }
  
    private void closeEngine(SSLEngine engine,
                             ByteBuffer netIn, ByteBuffer netOut,
                             ByteBuffer appIn, ByteBuffer appOut,
                             SocketChannel sc) throws IOException {
        SSLEngineResult result;
        SSLEngineResult.Status status;
        System.out.println(name + "closing engine..");  //do some cleanups
        engine.closeOutbound();
        //flush internal data out of the engine until there is no outbound data
        result = engine.wrap(appOut, netOut);
        netOut.flip();
        sc.write(netOut);
        netOut.clear();
        sc.read(netIn);
        netIn.flip();
        result = engine.unwrap(netIn, appIn);
        System.out.println(name + "inbound done: " + engine.isInboundDone());
        System.out.println(name + "outbound done: " + engine.isOutboundDone());
        System.out.println(name +  result.getStatus());
        sc.close();
    }

    private void doHandshake(SSLEngine engine, ByteBuffer netIn,
                             ByteBuffer netOut, ByteBuffer appIn,
                             ByteBuffer appOut, SocketChannel sc) throws IOException {
        SSLEngineResult result;
        HandshakeStatus status;

        loop: while ((status = engine.getHandshakeStatus()) != HandshakeStatus.NOT_HANDSHAKING) {
            System.out.println ("SSLClientEngine status " + status);
            switch (status) {
            case FINISHED:
                break loop;
            case NEED_WRAP:
                result = engine.wrap(appOut, netOut); //at handshake stage, data in appOut won't be processed
                if (result.bytesProduced() > 0) {
                    netOut.flip();
                    sc.write(netOut);
                    netOut.clear();
                }
                break;
            case NEED_TASK:
                Runnable task = engine.getDelegatedTask(); //these are the tasks like key generation that tend to take longer time to complete 
                if (task != null) {
                    task.run();  //it can be run at a different thread.
                } else {
                    status = HandshakeStatus.NEED_WRAP;
                }
                break;
            case NEED_UNWRAP:
                sc.read(netIn);
                netIn.flip();
                result = engine.unwrap(netIn, appIn); //at handshake stage, no data produced in appIn
                if (netIn.remaining() == 0) {
                    netIn.clear();
                } else { //if there are data left in the buffer, advance the buffer pointer so data won't be overwritten in next read() call
                    netIn.position(netIn.limit());
                    netIn.limit(netIn.capacity());
                }
            }
        }
        System.out.println(name + "handshake completes");
        SSLSession sess = engine.getSession();
        System.out.println(name + "connected using protocol " + sess.getProtocol());
        System.out.println(name + "connected using " + sess.getCipherSuite());
        System.out.println(name + "peer principal " + sess.getPeerPrincipal());
        reset(netIn, netOut, appIn, appOut);
    }

    private void getServerMessage(ByteBuffer appIn) {
        appIn.flip();
        byte[] mesg = new byte[appIn.remaining()];
        appIn.get(mesg);
        String message = new String(mesg);
        System.out.println("Server says: ");
        System.out.println(message);
    }

    private SSLEngine initEngine() throws IOException {
        SSLEngine engine = context.createSSLEngine(host, port); 
        engine.setUseClientMode(true);
        engine.beginHandshake();
        return engine;
    }

    private byte[] createMessage() {
        String message = "Hello from " + host + ", current time is " + new Date().toString();
        byte[] mesg = null;
        try {
            mesg = message.getBytes("UTF-8");
        } catch (java.io.UnsupportedEncodingException ex) {
            mesg = message.getBytes();
        }
        return mesg;
    }

    private void reset(ByteBuffer appIn, ByteBuffer appOut,
                      ByteBuffer netIn,
                      ByteBuffer netOut) {
       appIn.clear();
       appOut.clear();
       netIn.clear();
       netOut.clear();
   }

}
