/*
 * %Z%%M%    %I%  %W%  %G%  %U%
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * IBM Security Software Development Kit, Java (tm) Technology Edition
 *
 * (C) Copyright IBM Corp. 2005 All Rights Reserved.
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 * ===========================================================================
 */

import java.io.*;
import java.net.*;
import java.nio.*; 
import java.nio.channels.*;
import java.security.*;
import java.util.*;
import javax.net.ssl.*;
import javax.net.ssl.SSLEngineResult.*;

/**
 * SSLServerEngine
 *
 * Sample application to demonstrate a simple client/server interaction using SSLEngine. 
 * This is a single threaded server program that accepts one SSL connection each time. 
 * It requires client authentication by default. 
 * 
 * This sample program uses the JKS file called "testkeys" protected by the password 
 * "passphrase" as key store and trust store. At program start time, it loads the keys 
 * and certificates from the file to the memory and opens up a ServerSocketChannel. 
 * The channel is blocked on listening incoming message. For each SSL connection request, 
 * it creates an SSLEngine to handle the exchange of handshake and application data messages 
 * represented in ByteBuffer. Data transportation is handled by SocketChannel. For simplicity 
 * this program uses SocketChannel in blocking mode. Users can always explore the NIO 
 * (in java.nio etc package) functionalities and utilize non-blocking mode to build more 
 * sophisticated applications. When the engine receives a close notify message from the client, 
 * it will do some cleanup, then it will close the engine and socket channel, and the program 
 * will listen on the next incoming message. 
 *
 * Running the client/server application requires that the Server is started first and then the 
 * Client application.
 *
 * To start the Server program, execute the following command:
 *
 *     java SSLServerEngine
 *     or
 *     java SSLServerEngine <port number>
 *
 * The SSLServerEngine uses port 8050. Once it has started it waits for the client connection.
 *
 * To start the client, issue the following command:
 * 
 *     java SSLClientEngine <hostname> <port number>
 *     or
 *     java SSLClientEngine <hostname>
 *
 * where <hostname> is the name of the host running SSLServerEngine.  The default is localhost.
 * <port number> default is 8050. Both arguments are optional.
 *
 */

public class SSLServerEngine {
    private String name;
    private int port;
    private SSLContext context;
    private ServerSocketChannel ssc;
    private boolean clientAuth;

    SSLServerEngine(String name, int port) {
        this.name = name;
        this.port = port;
        initServerChannel(port);
    }

    private void initServerChannel(int port) {
        try {
            ssc = ServerSocketChannel.open();
            ServerSocket server_sock = ssc.socket();
            server_sock.bind(new InetSocketAddress((InetAddress)null, port));
        } catch (IOException ex) {
            System.out.println(ex.getMessage());
            System.exit(1);
        }
    }

    private void initContext(String keystore, char[] password) {
        try {
            KeyStore ks = KeyStore.getInstance("JKS");
            ks.load(new FileInputStream(keystore), password);  //load the keys and certificates
            KeyManagerFactory kmf = KeyManagerFactory.getInstance("IbmX509");
            kmf.init(ks, password);  //initialize the key manager factory
            TrustManagerFactory tmf =  TrustManagerFactory.getInstance("PKIX");
            tmf.init(ks); //initialize the trust manager factory
            context = SSLContext.getInstance("SSL_TLS");
            context.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null); //initialized the ssl context with key manager and trust manager
        } catch (IOException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (KeyStoreException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (NoSuchAlgorithmException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        } catch (Exception e) {
            System.out.println(e.getMessage());
            System.exit(1);
        }

    }

    public static void main(String args[]) {
        int port;
        if (args.length > 0) {
            port = Integer.parseInt(args[0]);
        }
        else port = 8050;
        SSLServerEngine server = new SSLServerEngine("SERVER: ", port);
        server.initContext("testkeys", "passphrase".toCharArray()); 
        boolean clientAuth = true;
        server.setClientAuthentication(clientAuth);  //it does client authentication
        try {
            server.run();
        } catch (IOException ex) {
            System.out.println(ex.getMessage());
            ex.printStackTrace();
        }
    }

    private void setClientAuthentication(boolean clientAuth) {
        this.clientAuth = clientAuth;
    }

    private void reset(ByteBuffer appIn, ByteBuffer appOut,
                       ByteBuffer netIn,
                       ByteBuffer netOut) {
        appIn.clear();
        appOut.clear();
        netIn.clear();
        netOut.clear();
    }

    private void run() throws IOException {
        SSLEngine engine = context.createSSLEngine();
        SocketChannel sc;
        int netBBSize = engine.getSession().getPacketBufferSize();
        int appBBSize = engine.getSession().getApplicationBufferSize();
        //allocate the max buffer size
        ByteBuffer netIn = ByteBuffer.allocate(netBBSize);
        ByteBuffer netOut = ByteBuffer.allocate(netBBSize);
        ByteBuffer appIn = ByteBuffer.allocate(appBBSize);
        ByteBuffer appOut = ByteBuffer.allocate(appBBSize);

        while (true) { //a single thread 
            System.out.println(name + "listening for connection request");
            sc = ssc.accept(); //block at the accept() call
            System.out.println(name + "accept a request");
            engine = initEngine();
            doHandshake(engine, netIn, netOut, appIn, appOut, sc); 
            processMessage(engine, netIn, netOut, appIn, appOut, sc);
            closeEngine(engine, netOut, appOut, sc);
        }

    }

    private void processMessage(SSLEngine engine, ByteBuffer netIn,
                                ByteBuffer netOut, ByteBuffer appIn, ByteBuffer appOut, SocketChannel sc) throws IOException {
        SSLEngineResult result;
        do {//keep accepting application data from the client, until it gets the close notify message
            System.out.println(name + "receiving data..");
            sc.read(netIn);
            netIn.flip();
            result = engine.unwrap(netIn, appIn);
            appIn.flip();

            if (result.getStatus() == Status.OK) {
                System.out.println("Client says:");
                byte[] received = new byte[appIn.remaining()];
                appIn.get(received);
                System.out.println(new String(received)); //print out the client message
                System.out.println(name + "sending data..");
                byte[] sent = new String("Received your message").getBytes(); 
                appOut.put(sent);
                appOut.flip();
                engine.wrap(appOut, netOut);
                netOut.flip();
                sc.write(netOut);  //send the receipt message
            }
            reset(netIn, netOut, appIn, appOut);
        }while (result.getStatus() != Status.CLOSED); //After processing the close notify message, engine status becomes CLOSED
    }

    private void closeEngine(SSLEngine engine,
                             ByteBuffer netOut, ByteBuffer appOut,
                             SocketChannel sc) throws IOException {
        SSLEngineResult result;
        SSLEngineResult.Status status;
    
        //flush internal data out of the engine
        result = engine.wrap(appOut, netOut);
        netOut.flip();
        sc.write(netOut);
        netOut.clear();
        
        if (!engine.isInboundDone()) { //engine no longer accepts inbound data. 
            engine.closeInbound();
        }
        System.out.println(name + "inbound done: " + engine.isInboundDone());
        System.out.println(name + "outbound done: " + engine.isOutboundDone());
        System.out.println(name +  result.getStatus());
        sc.close();
    }

    private void doHandshake(SSLEngine engine, ByteBuffer netIn,
                             ByteBuffer netOut, ByteBuffer appIn,
                             ByteBuffer appOut, SocketChannel sc) throws IOException {
        SSLEngineResult result;
        HandshakeStatus status;
        boolean hasData = false;

        loop: while ((status = engine.getHandshakeStatus()) != HandshakeStatus.NOT_HANDSHAKING) {
            System.out.println ("SSLServerEngine status " + status);
            switch (status) {
            case FINISHED:
                break loop;
            case NEED_WRAP:
                result = engine.wrap(appOut, netOut); //at handshake stage, data in appOut won't be processed
                if (result.bytesProduced() > 0) {
                    netOut.flip();
                    sc.write(netOut);
                    netOut.clear();
                }
                break;
            case NEED_TASK:
                Runnable task = engine.getDelegatedTask(); //these are the tasks like key generation that tend to take longer time to complete 
                if (task != null) {
                    task.run();  //it can be run at a different thread.
                } else {
                    status = HandshakeStatus.NEED_WRAP;
                }
                break;
            case NEED_UNWRAP:
                if (hasData == false) {
                    sc.read(netIn);
                    netIn.flip();

                }
                result = engine.unwrap(netIn, appIn); //at handshake stage, no data produced in appIn

                if (netIn.remaining() == 0) {
                    netIn.clear();
                    hasData = false;
                } else { //if there are data left in the buffer
                      hasData = true;
//                    netIn.position(netIn.limit());
//                    netIn.limit(netIn.capacity());
                }
            }
        }
        System.out.println(name + "handshake completes");
        SSLSession sess = engine.getSession();
        System.out.println(name + "connected using protocol " + sess.getProtocol());
        System.out.println(name + "connected using " + sess.getCipherSuite());
        System.out.println(name + "peer principal " + sess.getPeerPrincipal());
        reset(netIn, netOut, appIn, appOut);
    }

    private SSLEngine initEngine() throws IOException{
        //create the engine from the context. All engines created share the same ssl context.
        SSLEngine engine = context.createSSLEngine();         
        engine.setUseClientMode(false); //this is the server side engine
        engine.setNeedClientAuth(this.clientAuth); //do client authentication
        engine.beginHandshake(); //signal engine to get into handshake mode
        return engine;
    }
}
