/*
 * %Z%%M%    %I%  %W%  %G%  %U%
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 * IBM SDK, Java(tm) 2 Technology Edition, v5.0
 * (C) Copyright IBM Corp. 2004, 2005. All Rights Reserved
 * ===========================================================================
 */
import java.io.*;
import java.net.*;
import java.security.KeyStore;
import java.security.*;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.KeyManagerFactory;

/** Client/Server - ServerPKCS12KeyJsse 
  * <P>
  * This example shows a simple <I>Client/Server interaction</I> using the new JSSE2 provider.
  * For the sake of simplicity the ServerPKCS12KeyJsse and the ClientPKCS12KeyJsse (see ClientPKCS12KeyJsse.java) use
  * the same test PKCS12 KeyStore file testkeys.pfx. In real life there must be two different
  * PKCS12 KeyStore files, one for the client, and one for the server.
  * The PKCS12 KeyStore testkeys.pfx is protected by the password "passphrase".
  * When the ServerPKCS12KeyJsse is started it waits for the ClientPKCS12KeyJsse
  * connection. The ServerPKCS12KeyJsse uses the port 8050. To start the ClientPKCS12KeyJsse the host name
  * of the ServerPKCS12KeyJsse must be supplied (default: localhost).
  * First, ClientPKCS12KeyJsse and ServerPKCS12KeyJsse transmit a number of messages (N round-trips)
  * over several SSL connections; ClientPKCS12KeyJsse tries to connect to ServerPKCS12KeyJsse using all the
  * enabled cipher suites.
  * Than they establish a normal TCP connection and transmit the same number of
  * messages over it.
  */
public class ServerPKCS12KeyJsse {
    static int N = 50, L = 100, R = 2;

    ServerPKCS12KeyJsse() {
        super();
    }

    static SSLContext sslContext;

    private static void initContext()
    {
        try {

            if (Security.getProvider("com.ibm.jsse2.IBMJSSEProvider2") == null) {
                // IBMJSSE2 Provider is not installed, install it
                Security.addProvider(new com.ibm.jsse2.IBMJSSEProvider2());
            }

            KeyStore ks = KeyStore.getInstance("pkcs12");
            ks.load(new FileInputStream("testkeys.pfx"), "passphrase".toCharArray());
            KeyManagerFactory kmf = KeyManagerFactory.getInstance("NewIbmX509","IBMJSSE2");
            kmf.init(ks, "passphrase".toCharArray());
            TrustManagerFactory tmf = TrustManagerFactory.getInstance("X509","IBMJSSE2");
            tmf.init(ks);
            sslContext = SSLContext.getInstance("SSL_TLS","IBMJSSE2");
            sslContext.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null);
        } catch (Throwable t) {
            System.out.println("Failed to read testkeys.pfx file.");
            t.printStackTrace();
            System.exit(1);
        }
    }

/** Starts the ServerPKCS12KeyJsse application.
  * @param args no parameters required
  */
    public static void main(String args[]) {
        int i, j, len;

        if (args.length > 0)
            N = Integer.parseInt(args[0]);
        if (args.length > 1)
            L = Integer.parseInt(args[1]);

        byte buffer[] = new byte[L];

        System.out.println("SERVER: started.");

        initContext();

        try {

            System.out.println("ServerPKCS12KeyJsse: SSL server context created.");
            System.out.println("ServerPKCS12KeyJsse: enabled cipher suites");
            String enabled[] = sslContext.getSocketFactory().getSupportedCipherSuites();
            for (i=0; i < enabled.length; i++)
                System.out.println(enabled[i]);

            SSLServerSocketFactory factory = sslContext.getServerSocketFactory();
            SSLServerSocket ssl_server_sock = (SSLServerSocket)factory.createServerSocket(8050);
            SSLSocket ssl_sock; 
            InputStream istr;
            OutputStream ostr;
            long t;
            while (buffer[0] != -1) {
                for (int n=0; n<R; n++) {
                    try {
                        ssl_sock = (SSLSocket)(ssl_server_sock.accept());
                        ssl_sock.setEnabledCipherSuites(enabled);

                        ssl_sock.startHandshake();
                        SSLSession session = ssl_sock.getSession();
                        System.out.println("\nServerPKCS12KeyJsse: SSL connection established");
                        System.out.println("   cipher suite:       " + session.getCipherSuite());
                    } catch (IOException se) {
                        System.out.println("\nServerPKCS12KeyJsse: client connection refused\n" + se);
                        se.printStackTrace();
                        break;
                    }

                    if (L > 0) {
                        istr = ssl_sock.getInputStream();
                        ostr = ssl_sock.getOutputStream();

                        t = System.currentTimeMillis(); 
                        for (j=0;j<N;j++) {
                            for (len = 0;;) {
                                try {
                                    if ((i = istr.read(buffer, len, L-len)) == -1) {
                                        System.out.println("ServerPKCS12KeyJsse: connection dropped by partner.");
                                        ssl_sock.close();
                                        return;
                                    }
                                } catch (InterruptedIOException e) {
                                    System.out.println("waiting");
                                    continue;
                                }
                                // System.out.println("ServerPKCS12KeyJsse: " + i + " bytes received.");
                                if ((len+=i) == L) break;
                            }
                            ostr.write(buffer, 0, L);
                        }
                        System.out.println("Messages = " + N*2 + "; Time = " +
                                           (System.currentTimeMillis() - t));
                    }
                    ssl_sock.close();
                }
            }
            ssl_server_sock.close();
            System.out.println("\nServerPKCS12KeyJsse: SSL connection closed.");
            Thread.sleep(1000);

            // plain socket benchmark
            if (L > 0) {
                // Socket bench
                ServerSocket server_sock = new ServerSocket(8050);
                Socket sock = server_sock.accept();
                System.out.println("\nServerPKCS12KeyJsse: Socket connection accepted.");
                istr = sock.getInputStream();
                ostr = sock.getOutputStream();

                t = System.currentTimeMillis();
                for (j=0;j<N;j++) {
                    for (len = 0;;) {
                        if ((i = istr.read(buffer, len, L-len)) == -1) {
                            System.out.println("ServerPKCS12KeyJsse: connection dropped by partner.");
                            return;
                        }
                        // System.out.println("ServerPKCS12KeyJsse: " + i + " bytes received.");
                        if ((len+=i) == L) break;
                    }
                    ostr.write(buffer, 0, L);
                }
                System.out.println("Messages = " + N*2 + "; Time = " +
                                   (System.currentTimeMillis() - t));
                sock.close();
                server_sock.close();
                System.out.println("\nServerPKCS12KeyJsse: Socket connection closed.");
                Thread.sleep(1000);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("ServerPKCS12KeyJsse: terminated.");
    }
}
